from __future__ import annotations

import os
import sys


def _py_abspath(path: str) -> str:
    """
    special version of abspath
    that will leave paths from jython jars alone
    """
    if path.startswith("__pyclasspath__"):
        return path
    else:
        return os.path.abspath(path)


def distribution_version(name: str) -> str | None:
    """try to get the version of the named distribution,
    returns None on failure"""
    if sys.version_info >= (3, 8):
        from importlib.metadata import PackageNotFoundError, version
    else:
        from importlib_metadata import PackageNotFoundError, version
    try:
        return version(name)
    except PackageNotFoundError:
        return None


def importobj(modpath: str, attrname: str | None) -> object:
    """imports a module, then resolves the attrname on it"""
    module = __import__(modpath, None, None, ["__doc__"])
    if not attrname:
        return module

    retval = module
    names = attrname.split(".")
    for x in names:
        retval = getattr(retval, x)
    return retval



"""
"""
import os
import shutil
import warnings
from logging import getLogger
from pathlib import Path
from subprocess import CompletedProcess
from tempfile import NamedTemporaryFile
from typing import Any, Dict, Optional, Tuple

from ..utils import WinLex, logged_run, unlink
from .base import Menu, MenuItem
from .win_utils.knownfolders import folder_path as windows_folder_path
from .win_utils.registry import (
    register_file_extension,
    register_url_protocol,
    unregister_file_extension,
    unregister_url_protocol,
)

log = getLogger(__name__)


class WindowsMenu(Menu):
    def create(self) -> Tuple[os.PathLike]:
        log.debug("Creating %s", self.start_menu_location)
        self.start_menu_location.mkdir(parents=True, exist_ok=True)
        if self.quick_launch_location:
            self.quick_launch_location.mkdir(parents=True, exist_ok=True)
        if self.desktop_location:
            self.desktop_location.mkdir(parents=True, exist_ok=True)
        return (self.start_menu_location,)

    def remove(self) -> Tuple[os.PathLike]:
        log.debug("Removing %s", self.start_menu_location)
        shutil.rmtree(self.start_menu_location, ignore_errors=True)
        return (self.start_menu_location,)

    @property
    def start_menu_location(self) -> Path:
        """
        On Windows we can create shortcuts in several places:

        - Start Menu
        - Desktop
        - Quick launch (only for user installs)

        In this property we only report the path to the Start menu.
        For other menus, check their respective properties.
        """
        return Path(windows_folder_path(self.mode, False, "start")) / self.name

    @property
    def quick_launch_location(self) -> Path:
        if self.mode == "system":
            # TODO: Check if this is true?
            warnings.warn("Quick launch menus are not available for system level installs")
            return
        return Path(windows_folder_path(self.mode, False, "quicklaunch"))

    @property
    def desktop_location(self) -> Path:
        return Path(windows_folder_path(self.mode, False, "desktop"))

    @property
    def placeholders(self) -> Dict[str, str]:
        placeholders = super().placeholders
        placeholders.update(
            {
                "SCRIPTS_DIR": str(self.prefix / "Scripts"),
                "PYTHON": str(self.prefix / "python.exe"),
                "PYTHONW": str(self.prefix / "pythonw.exe"),
                "BASE_PYTHON": str(self.base_prefix / "python.exe"),
                "BASE_PYTHONW": str(self.base_prefix / "pythonw.exe"),
                "BIN_DIR": str(self.prefix / "Library" / "bin"),
                "SP_DIR": str(self._site_packages()),
                "ICON_EXT": "ico",
            }
        )
        return placeholders

    def _conda_exe_path_candidates(self):
        return (
            self.base_prefix / "_conda.exe",
            Path(os.environ.get("CONDA_EXE", r"C:\oops\a_file_that_does_not_exist")),
            self.base_prefix / "condabin" / "conda.bat",
            self.base_prefix / "bin" / "conda.bat",
            Path(os.environ.get("MAMBA_EXE", r"C:\oops\a_file_that_does_not_exist")),
            self.base_prefix / "condabin" / "micromamba.exe",
            self.base_prefix / "bin" / "micromamba.exe",
        )

    def render(
        self, value: Any, slug: bool = False, extra: Optional[Dict[str, str]] = None
    ) -> Any:
        """
        We extend the render method here to replace forward slashes with backslashes.
        We ONLY do it if the string does not start with /, because it might
        be just a windows-style command-line flag (e.g. cmd /K something).

        This will not escape strings such as `/flag:something`, common
        in compiler stuff but we can assume such windows specific
        constructs will have their platform override, which is always an option.

        This is just a convenience for things like icon paths or Unix-like paths
        in the command key.
        """
        value = super().render(value, slug=slug, extra=extra)
        if hasattr(value, "replace") and "/" in value and value[0] != "/":
            value = value.replace("/", "\\")
        return value

    def _site_packages(self, prefix=None) -> Path:
        if prefix is None:
            prefix = self.prefix
        return self.prefix / "Lib" / "site-packages"

    def _paths(self) -> Tuple[Path]:
        return (self.start_menu_location,)


class WindowsMenuItem(MenuItem):
    @property
    def location(self) -> Path:
        """
        Path to the .lnk file placed in the Start Menu
        On Windows, menuinst can create up to three shortcuts (start menu, desktop, quick launch)
        This property only lists the one for start menu
        """
        return self.menu.start_menu_location / self._shortcut_filename()

    def create(self) -> Tuple[Path, ...]:
        from .win_utils.winshortcut import create_shortcut

        self._precreate()
        paths = self._paths()

        for path in paths:
            if not path.suffix == ".lnk":
                continue

            target_path, *arguments = self._process_command()
            working_dir = self.render_key("working_dir")
            if working_dir:
                Path(working_dir).mkdir(parents=True, exist_ok=True)
            else:
                working_dir = "%HOMEPATH%"

            icon = self.render_key("icon") or ""

            # winshortcut is a windows-only C extension! create_shortcut has this API
            # Notice args must be passed as positional, no keywords allowed!
            # winshortcut.create_shortcut(path, description, filename, arguments="",
            #                             workdir=None, iconpath=None, iconindex=0, app_id="")
            create_shortcut(
                target_path,
                self._shortcut_filename(ext=""),
                str(path),
                " ".join(arguments),
                working_dir,
                icon,
                0,
                self._app_user_model_id(),
            )

        self._register_file_extensions()
        self._register_url_protocols()

        return paths

    def remove(self) -> Tuple[Path, ...]:
        self._unregister_file_extensions()
        self._unregister_url_protocols()

        paths = self._paths()
        for path in paths:
            log.debug("Removing %s", path)
            unlink(path, missing_ok=True)

        return paths

    def _paths(self) -> Tuple[Path, ...]:
        paths = [self.location]
        extra_dirs = []
        if self.metadata["desktop"]:
            extra_dirs.append(self.menu.desktop_location)
        if self.metadata["quicklaunch"] and self.menu.quick_launch_location:
            extra_dirs.append(self.menu.quick_launch_location)

        if extra_dirs:
            paths += [directory / self._shortcut_filename() for directory in extra_dirs]

        if self.metadata["activate"]:
            # This is the accessory launcher script for cmd
            paths.append(self._path_for_script())

        return tuple(paths)

    def _shortcut_filename(self, ext: str = "lnk"):
        env_suffix = f" ({self.menu.env_name})" if self.menu.env_name else ""
        ext = f".{ext}" if ext else ""
        return f"{self.render_key('name', extra={})}{env_suffix}{ext}"

    def _path_for_script(self) -> Path:
        return Path(self.menu.placeholders["MENU_DIR"]) / self._shortcut_filename("bat")

    def _precreate(self):
        precreate_code = self.render_key("precreate")
        if not precreate_code:
            return
        with NamedTemporaryFile(delete=False, mode="w") as tmp:
            tmp.write(precreate_code)
        system32 = Path(os.environ.get("SystemRoot", "C:\\Windows")) / "system32"
        cmd = [
            str(system32 / "WindowsPowerShell" / "v1.0" / "powershell.exe"),
            f"\"start '{tmp.name}' -WindowStyle hidden\"",
        ]
        logged_run(cmd, check=True)
        os.unlink(tmp.name)

    def _command(self) -> str:
        lines = [
            "@ECHO OFF",
            ":: Script generated by conda/menuinst",
        ]
        precommand = self.render_key("precommand")
        if precommand:
            lines.append(precommand)
        if self.metadata["activate"]:
            conda_exe = self.menu.conda_exe
            if self.menu._is_micromamba(conda_exe):
                activate = "shell activate"
            else:
                activate = "shell.cmd.exe activate"
            activator = f'{self.menu.conda_exe} {activate} "{self.menu.prefix}"'
            lines += [
                "@SETLOCAL ENABLEDELAYEDEXPANSION",
                f'@FOR /F "usebackq tokens=*" %%i IN (`{activator}`) do set "ACTIVATOR=%%i"',
                "@CALL %ACTIVATOR%",
                ":: This below is the user command",
            ]

        lines.append(" ".join(WinLex.quote_args(self.render_key("command"))))

        return "\r\n".join(lines)

    def _write_script(self, script_path: Optional[os.PathLike] = None) -> Path:
        """
        This method generates the batch script that will be called by the shortcut
        """
        if script_path is None:
            script_path = self._path_for_script()
        else:
            script_path = Path(script_path)

        script_path.parent.mkdir(parents=True, exist_ok=True)
        with open(script_path, "w") as f:
            f.write(self._command())

        return script_path

    def _process_command(self, with_arg1=False) -> Tuple[str]:
        if self.metadata["activate"]:
            script = self._write_script()
            if self.metadata["terminal"]:
                command = ["cmd", "/D", "/K", f'"{script}"']
                if with_arg1:
                    command.append("%1")
            else:
                system32 = Path(os.environ.get("SystemRoot", "C:\\Windows")) / "system32"
                arg1 = "%1 " if with_arg1 else ""
                # This is an UGLY hack to start the script in a hidden window
                # We use CMD to call PowerShell to call the BAT file
                # This flashes faster than Powershell -> BAT! Don't ask me why.
                command = [
                    f'"{system32 / "cmd.exe"}"',
                    "/D",
                    "/C",
                    "START",
                    "/MIN",
                    '""',
                    f'"{system32 / "WindowsPowerShell" / "v1.0" / "powershell.exe"}"',
                    "-WindowStyle",
                    "hidden",
                    f"\"start '{script}' {arg1}-WindowStyle hidden\"",
                ]
            return command

        command = self.render_key("command")
        if with_arg1 and all("%1" not in arg for arg in command):
            command.append("%1")
        return WinLex.quote_args(command)

    def _ftype_identifier(self, extension):
        identifier = self.render_key("name", slug=True)
        return f"{identifier}.AssocFile{extension}"

    def _register_file_extensions_cmd(self):
        """
        This function uses CMD's `assoc` and `ftype` commands.
        """
        extensions = self.metadata["file_extensions"]
        if not extensions:
            return
        command = " ".join(self._process_command())
        exts = list(dict.fromkeys([ext.lower() for ext in extensions]))
        for ext in exts:
            identifier = self._ftype_identifier(ext)
            self._cmd_ftype(identifier, command)
            self._cmd_assoc(ext, associate_to=identifier)

    def _unregister_file_extensions_cmd(self):
        """
        This function uses CMD's `assoc` and `ftype` commands.
        """
        extensions = self.metadata["file_extensions"]
        if not extensions:
            return
        exts = list(dict.fromkeys([ext.lower() for ext in extensions]))
        for ext in exts:
            identifier = self._ftype_identifier(ext)
            self._cmd_ftype(identifier)  # remove
            # TODO: Do we need to clean up the `assoc` mappings too?

    @staticmethod
    def _cmd_assoc(extension, associate_to=None, query=False, remove=False) -> CompletedProcess:
        "https://learn.microsoft.com/en-us/windows-server/administration/windows-commands/assoc"
        if sum([associate_to, query, remove]) != 1:
            raise ValueError("Only one of {associate_to, query, remove} must be set.")
        if not extension.startswith("."):
            raise ValueError("extension must startwith '.'")
        if associate_to:
            arg = f"{extension}={associate_to}"
        elif query:
            arg = extension
        elif remove:
            arg = f"{extension}="
        return logged_run(["cmd", "/D", "/C", f"assoc {arg}"], check=True)

    @staticmethod
    def _cmd_ftype(identifier, command=None, query=False, remove=False) -> CompletedProcess:
        "https://learn.microsoft.com/en-us/windows-server/administration/windows-commands/ftype"
        if sum([command, query, remove]) != 1:
            raise ValueError("Only one of {command, query, remove} must be set.")
        if command:
            arg = f"{identifier}={command}"
        elif query:
            arg = identifier
        elif remove:
            arg = f"{identifier}="
        return logged_run(["cmd", "/D", "/C", f"assoc {arg}"], check=True)

    def _register_file_extensions(self):
        """WIP"""
        extensions = self.metadata["file_extensions"]
        if not extensions:
            return

        command = " ".join(self._process_command(with_arg1=True))
        icon = self.render_key("icon")
        exts = list(dict.fromkeys([ext.lower() for ext in extensions]))
        for ext in exts:
            identifier = self._ftype_identifier(ext)
            register_file_extension(ext, identifier, command, icon=icon, mode=self.menu.mode)

    def _unregister_file_extensions(self):
        extensions = self.metadata["file_extensions"]
        if not extensions:
            return

        exts = list(dict.fromkeys([ext.lower() for ext in extensions]))
        for ext in exts:
            identifier = self._ftype_identifier(ext)
            unregister_file_extension(ext, identifier, mode=self.menu.mode)

    def _register_url_protocols(self):
        "See https://learn.microsoft.com/en-us/previous-versions/windows/internet-explorer/ie-developer/platform-apis/aa767914(v=vs.85)"  # noqa
        protocols = self.metadata["url_protocols"]
        if not protocols:
            return
        command = " ".join(self._process_command(with_arg1=True))
        icon = self.render_key("icon")
        for protocol in protocols:
            identifier = self._ftype_identifier(protocol)
            register_url_protocol(protocol, command, identifier, icon=icon, mode=self.menu.mode)

    def _unregister_url_protocols(self):
        protocols = self.metadata["url_protocols"]
        if not protocols:
            return
        for protocol in protocols:
            identifier = self._ftype_identifier(protocol)
            unregister_url_protocol(protocol, identifier, mode=self.menu.mode)

    def _app_user_model_id(self):
        aumi = self.render_key("app_user_model_id")
        if not aumi:
            return f"Menuinst.{self.render_key('name', slug=True).replace('.', '')}"[:128]
        return aumi



# this script is used on windows to wrap shortcuts so that they are executed within an environment
#   It only sets the appropriate prefix PATH entries - it does not actually activate environments

import argparse
import os
import subprocess
import sys
from os.path import join, pathsep

from menuinst._legacy.knownfolders import FOLDERID, get_folder_path

# call as: python cwp.py [--no-console] PREFIX ARGs...
parser = argparse.ArgumentParser()
parser.add_argument(
    "--no-console", action="store_true", help="Create subprocess with CREATE_NO_WINDOW flag."
)
parser.add_argument("prefix", help="Prefix to be 'activated' before calling `args`.")
parser.add_argument("args", nargs="*", help="Command (and arguments) to be executed.")
parsed_args = parser.parse_args()

no_console = parsed_args.no_console
prefix = parsed_args.prefix
args = parsed_args.args

new_paths = pathsep.join(
    [
        prefix,
        join(prefix, "Library", "mingw-w64", "bin"),
        join(prefix, "Library", "usr", "bin"),
        join(prefix, "Library", "bin"),
        join(prefix, "Scripts"),
    ]
)
env = os.environ.copy()
env["PATH"] = new_paths + pathsep + env["PATH"]
env["CONDA_PREFIX"] = prefix

documents_folder, exception = get_folder_path(FOLDERID.Documents)
if exception:
    documents_folder, exception = get_folder_path(FOLDERID.PublicDocuments)
if not exception:
    os.chdir(documents_folder)

creationflags = {}
if no_console:
    creationflags["creationflags"] = getattr(subprocess, "CREATE_NO_WINDOW", 0x08000000)
sys.exit(subprocess.call(args, env=env, **creationflags))



"""
"""
import os
import platform
import plistlib
import shutil
from hashlib import sha1
from logging import getLogger
from pathlib import Path
from typing import Dict, Optional, Tuple

from .. import data as _menuinst_data
from ..utils import UnixLex, logged_run
from .base import Menu, MenuItem, menuitem_defaults

log = getLogger(__name__)


class MacOSMenu(Menu):
    def create(self) -> Tuple:
        return self._paths()

    def remove(self) -> Tuple:
        return self._paths()

    @property
    def placeholders(self) -> Dict[str, str]:
        placeholders = super().placeholders
        placeholders.update(
            {
                "SP_DIR": str(self._site_packages()),
                "ICON_EXT": "icns",
                "PYTHONAPP": str(self.prefix / "python.app" / "Contents" / "MacOS" / "python"),
            }
        )
        return placeholders

    def _paths(self) -> Tuple:
        return ()


class MacOSMenuItem(MenuItem):
    @property
    def location(self) -> Path:
        "Path to the .app directory defining the menu item"
        return self._base_location() / "Applications" / self._bundle_name

    @property
    def _bundle_name(self) -> str:
        return f"{self.render_key('name', extra={})}.app"

    @property
    def _nested_location(self) -> Path:
        "Path to the nested .app directory defining the menu item main app"
        return self.location / "Contents" / "Resources" / self._bundle_name

    def _base_location(self) -> Path:
        if self.menu.mode == "user":
            return Path("~").expanduser()
        return Path("/")

    def _precreate(self):
        super()._precreate()
        for src, dest in (self.metadata["link_in_bundle"] or {}).items():
            rendered_dest: Path = (self.location / self.render(dest)).resolve()
            # if not rendered_dest.is_relative_to(self.location):  # FUTURE: Only for 3.9+
            if not str(rendered_dest).startswith(str(self.location)):
                raise ValueError(
                    "'link_in_bundle' destinations MUST be created "
                    f"inside the .app bundle ({self.location}), but it points to '{rendered_dest}."
                )
            rendered_dest.parent.mkdir(parents=True, exist_ok=True)
            os.symlink(self.render(src), rendered_dest)

    def create(self) -> Tuple[Path]:
        log.debug("Creating %s", self.location)
        self._create_application_tree()
        self._precreate()
        self._copy_icon()
        self._write_pkginfo()
        self._write_plistinfo()
        self._write_appkit_launcher()
        self._write_launcher()
        self._write_script()
        self._write_event_handler()
        self._maybe_register_with_launchservices()
        self._sign_with_entitlements()
        return (self.location,)

    def remove(self) -> Tuple[Path]:
        log.debug("Removing %s", self.location)
        self._maybe_register_with_launchservices(register=False)
        shutil.rmtree(self.location, ignore_errors=True)
        return (self.location,)

    def _create_application_tree(self) -> Tuple[Path]:
        paths = [
            self.location / "Contents" / "Resources",
            self.location / "Contents" / "MacOS",
        ]
        if self._needs_appkit_launcher:
            paths += [
                self._nested_location / "Contents" / "Resources",
                self._nested_location / "Contents" / "MacOS",
            ]
        for path in paths:
            path.mkdir(parents=True, exist_ok=False)
        return tuple(paths)

    def _copy_icon(self):
        icon = self.render_key("icon")
        if icon:
            shutil.copy(icon, self.location / "Contents" / "Resources")
            if self._needs_appkit_launcher:
                shutil.copy(icon, self._nested_location / "Contents" / "Resources")

    def _write_pkginfo(self):
        app_bundles = [self.location]
        if self._needs_appkit_launcher:
            app_bundles.append(self._nested_location)
        for app in app_bundles:
            with open(app / "Contents" / "PkgInfo", "w") as f:
                f.write(f"APPL{self.render_key('name', slug=True)[:8]}")

    def _write_plistinfo(self):
        name = self.render_key("name")
        slugname = self.render_key("name", slug=True)
        if len(slugname) > 16:
            shortname = slugname[:10] + sha1(slugname.encode()).hexdigest()[:6]
        else:
            shortname = slugname
        pl = {
            "CFBundleName": shortname,
            "CFBundleDisplayName": name,
            "CFBundleExecutable": slugname,
            "CFBundleGetInfoString": f"{slugname}-1.0.0",
            "CFBundleIdentifier": f"com.{slugname}",
            "CFBundlePackageType": "APPL",
            "CFBundleVersion": "1.0.0",
            "CFBundleShortVersionString": "1.0.0",
        }

        icon = self.render_key("icon")
        if icon:
            pl["CFBundleIconFile"] = Path(icon).name

        if self._needs_appkit_launcher:
            # write only the basic plist info into the nested bundle
            with open(self._nested_location / "Contents" / "Info.plist", "wb") as f:
                plistlib.dump(pl, f)
            # the *outer* bundle is background-only and needs a different ID
            pl["LSBackgroundOnly"] = True
            pl["CFBundleIdentifier"] = f"com.{slugname}-appkit-launcher"

        # Override defaults with (potentially) user provided values
        ignore_keys = (*menuitem_defaults, "entitlements", "link_in_bundle")
        for key in menuitem_defaults["platforms"]["osx"]:
            if key in ignore_keys:
                continue
            value = self.render_key(key)
            if value is None:
                continue
            if key == "CFBundleVersion":
                # setting the version also changes these two values
                pl["CFBundleShortVersionString"] = value
                pl["CFBundleGetInfoString"] = f"{slugname}-{value}"
            pl[key] = value
        with open(self.location / "Contents" / "Info.plist", "wb") as f:
            plistlib.dump(pl, f)

    def _command(self) -> str:
        lines = ["#!/bin/sh"]
        if self.render_key("terminal"):
            # FIXME: Terminal launching will miss the arguments;
            # there's no easy way to pass them!
            lines.extend(
                [
                    'if [ "${__CFBundleIdentifier:-}" != "com.apple.Terminal" ]; then',
                    '    open -b com.apple.terminal "$0"',
                    '    exit $?',
                    'fi',
                ]
            )

        working_dir = self.render_key("working_dir")
        if working_dir:
            Path(working_dir).mkdir(parents=True, exist_ok=True)
            lines.append(f'cd "{working_dir}"')

        precommand = self.render_key("precommand")
        if precommand:
            lines.append(precommand)

        if self.metadata["activate"]:
            conda_exe = self.menu.conda_exe
            if self.menu._is_micromamba(conda_exe):
                activate = "shell activate"
            else:
                activate = "shell.bash activate"
            lines.append(f'eval "$("{conda_exe}" {activate} "{self.menu.prefix}")"')

        lines.append(" ".join(UnixLex.quote_args(self.render_key("command"))))

        return "\n".join(lines)

    def _write_appkit_launcher(self, launcher_path: Optional[os.PathLike] = None) -> os.PathLike:
        if launcher_path is None:
            launcher_path = self._default_appkit_launcher_path()
        shutil.copy(self._find_appkit_launcher(), launcher_path)
        os.chmod(launcher_path, 0o755)
        return launcher_path

    def _write_launcher(self, launcher_path: Optional[os.PathLike] = None) -> os.PathLike:
        if launcher_path is None:
            launcher_path = self._default_launcher_path()
        shutil.copy(self._find_launcher(), launcher_path)
        os.chmod(launcher_path, 0o755)
        return launcher_path

    def _write_script(self, script_path: Optional[os.PathLike] = None) -> os.PathLike:
        if script_path is None:
            script_path = self._default_launcher_path(suffix="-script")
        with open(script_path, "w") as f:
            f.write(self._command())
        os.chmod(script_path, 0o755)
        return script_path

    def _write_event_handler(self, script_path: Optional[os.PathLike] = None) -> os.PathLike:
        if not self._needs_appkit_launcher:
            return
        event_handler_logic = self.render_key("event_handler")
        if event_handler_logic is None:
            return
        if script_path is None:
            script_path = self.location / "Contents" / "Resources" / "handle-event"
        with open(script_path, "w") as f:
            f.write(f"#!/bin/bash\n{event_handler_logic}\n")
        os.chmod(script_path, 0o755)
        return script_path

    def _paths(self) -> Tuple[os.PathLike]:
        return (self.location,)

    def _find_appkit_launcher(self) -> Path:
        launcher_name = f"appkit_launcher_{platform.machine()}"
        for datapath in _menuinst_data.__path__:
            launcher_path = Path(datapath) / launcher_name
            if launcher_path.is_file() and os.access(launcher_path, os.X_OK):
                return launcher_path
        raise ValueError(f"Could not find executable launcher for {platform.machine()}")

    def _find_launcher(self) -> Path:
        launcher_name = f"osx_launcher_{platform.machine()}"
        for datapath in _menuinst_data.__path__:
            launcher_path = Path(datapath) / launcher_name
            if launcher_path.is_file() and os.access(launcher_path, os.X_OK):
                return launcher_path
        raise ValueError(f"Could not find executable launcher for {platform.machine()}")

    def _default_appkit_launcher_path(self, suffix: str = "") -> Path:
        name = self.render_key("name", slug=True)
        return self.location / "Contents" / "MacOS" / f'{name}{suffix}'

    def _default_launcher_path(self, suffix: str = "") -> Path:
        name = self.render_key("name", slug=True)
        if self._needs_appkit_launcher:
            return self._nested_location / "Contents" / "MacOS" / f'{name}{suffix}'
        return self.location / "Contents" / "MacOS" / f'{name}{suffix}'

    def _maybe_register_with_launchservices(self, register=True):
        if not self._needs_appkit_launcher:
            return
        if register:
            # register the URL scheme with `lsregister`
            _lsregister("-R", str(self.location))
        else:
            _lsregister("-R", "-u", "-all", str(self.location))

    def _sign_with_entitlements(self):
        "Self-sign shortcut to apply required entitlements"
        entitlement_keys = self.render_key("entitlements")
        if not entitlement_keys:
            return
        slugname = self.render_key("name", slug=True)
        plist = {key: True for key in entitlement_keys}
        entitlements_path = self.location / "Contents" / "Entitlements.plist"
        with open(entitlements_path, "wb") as f:
            plistlib.dump(plist, f)
        logged_run(
            [
                # hardcode to system location to avoid accidental clobber in PATH
                "/usr/bin/codesign",
                "--verbose",
                "--sign",
                "-",
                "--prefix",
                f"com.{slugname}",
                "--options",
                "runtime",
                "--force",
                "--deep",
                "--entitlements",
                entitlements_path,
                self.location,
            ],
            check=True,
        )

    @property
    def _needs_appkit_launcher(self) -> bool:
        """
        In macOS, file type and URL protocol associations are handled by the
        Apple Events system. When the user opens on a file or URL, the system
        will send an Apple Event to the application that was registered as a handler.
        We need a special launcher to handle these events and pass them to the
        wrapped application in the shortcut.

        See:
        - https://developer.apple.com/library/archive/documentation/Carbon/Conceptual/LaunchServicesConcepts/LSCConcepts/LSCConcepts.html  # noqa
        - The source code at /src/appkit-launcher in this repository
        """
        needed_keys = ("CFBundleURLTypes", "CFBundleDocumentTypes")
        return any([self.metadata.get(k) for k in needed_keys])


def _lsregister(*args, check=True, **kwargs):
    exe = (
        "/System/Library/Frameworks/CoreServices.framework"
        "/Frameworks/LaunchServices.framework/Support/lsregister"
    )
    return logged_run([exe, *args], check=check, **kwargs)



"""
Generate JSON schemas from pydantic models
"""

# flake8: noqa

import json
from logging import getLogger
from pathlib import Path
from pprint import pprint
from typing import Dict, List, Literal, Optional, Union

from pydantic import BaseModel as _BaseModel
from pydantic import Field, conlist, constr

log = getLogger(__name__)


class BaseModel(_BaseModel):
    class Config:
        extra = "forbid"


class BasePlatformSpecific(BaseModel):
    """
    Same as :class:`MenuItem`, but without ``platforms``, and all is optional.

    Note:
    * This needs to be kept up-to-date with :class:`MenuItem`!
    * Default value is always ``None``.
    """

    name: Optional[constr(min_length=1)] = None
    "The name of the menu item"
    description: Optional[str] = None
    "A longer description of the menu item. Shown on popup messages."
    icon: Optional[constr(min_length=1)] = None
    "Path to the file representing or containing the icon."
    command: Optional[conlist(str, min_items=1)] = None
    """
    Command to run with the menu item, expressed as a
    list of strings where each string is an argument.
    """
    working_dir: Optional[constr(min_length=1)] = None
    """
    Working directory for the running process.
    Defaults to user directory on each platform.
    """
    precommand: Optional[constr(min_length=1)] = None
    """
    (Simple, preferrably single-line) logic to run before the command is run.
    Runs before the env is activated, if that applies.
    """
    precreate: Optional[constr(min_length=1)] = None
    "(Simple, preferrably single-line) logic to run before the shortcut is created."
    activate: Optional[bool] = None
    "Whether to activate the target environment before running ``command``."
    terminal: Optional[bool] = None
    """
    Whether run the program in a terminal/console or not.
    On Windows, it only has an effect if ``activate`` is true.
    On MacOS, the application will ignore command-line arguments.
    """


class Windows(BasePlatformSpecific):
    "Windows-specific instructions. You can override global keys here if needed"

    desktop: Optional[bool] = True
    "Whether to create a desktop icon in addition to the Start Menu item."
    quicklaunch: Optional[bool] = True
    "Whether to create a quick launch icon in addition to the Start Menu item."
    url_protocols: Optional[List[constr(regex=r"\S+")]] = None
    "URL protocols that will be associated with this program."
    file_extensions: Optional[List[constr(regex=r"\.\S*")]] = None
    "File extensions that will be associated with this program."
    app_user_model_id: Optional[constr(regex=r"\S+\.\S+", max_length=128)] = None
    """
    Identifier used in Windows 7 and above to associate processes, files and windows with a
    particular application. If your shortcut produces duplicated icons, you need to define this
    field. If not set, it will default to ``Menuinst.<name>``.

    See `AppUserModelID docs <aumi-docs>`__ for more information on the required string format.

    .. aumi-docs: https://learn.microsoft.com/en-us/windows/win32/shell/appids#how-to-form-an-application-defined-appusermodelid
    """


class Linux(BasePlatformSpecific):
    """
    Linux-specific instructions.

    Check the `Desktop entry specification <desktop-entry-spec>`__ for more details.

    .. desktop-entry-spec: https://specifications.freedesktop.org/desktop-entry-spec/desktop-entry-spec-latest.html#recognized-keys
    """

    Categories: Optional[Union[List[str], constr(regex=r"^.+;$")]] = None
    """
    Categories in which the entry should be shown in a menu.
    "See 'Registered categories' in the `Menu Spec <menu-spec>`__.

    .. menu-spec: http://www.freedesktop.org/Standards/menu-spec
    """
    DBusActivatable: Optional[bool] = None
    "A boolean value specifying if D-Bus activation is supported for this application."
    GenericName: Optional[str] = None
    """
    Generic name of the application; e.g. if the name is 'conda',
    this would be 'Package Manager'.
    """
    Hidden: Optional[bool] = None
    "Disable shortcut, signaling a missing resource."
    Implements: Optional[Union[List[str], constr(regex=r"^.+;$")]] = None
    """
    List of supported interfaces. See 'Interfaces' in `Desktop Entry Spec <desktop-entry-spec>`__.

    .. desktop-entry-spec: https://specifications.freedesktop.org/desktop-entry-spec/desktop-entry-spec-latest.html#interfaces
    """
    Keywords: Optional[Union[List[str], constr(regex=r"^.+;$")]] = None
    "Additional terms to describe this shortcut to aid in searching."
    MimeType: Optional[Union[List[str], constr(regex=r"^.+;$")]] = None
    """
    The MIME type(s) supported by this application.
    Note this includes file types and URL protocols.
    For URL protocols, use ``x-scheme-handler/your-protocol-here``.
    For example, if you want to register ``menuinst:``, you would
    include ``x-scheme-handler/menuinst``.
    """
    NoDisplay: Optional[bool] = None
    """
    Do not show this item in the menu. Useful to associate MIME types
    and other registrations, without having an actual clickable item.
    Not to be confused with 'Hidden'.
    """
    NotShowIn: Optional[Union[List[str], constr(regex=r"^.+;$")]] = None
    """
    Desktop environments that should NOT display this item.
    It'll check against ``$XDG_CURRENT_DESKTOP``."
    """
    OnlyShowIn: Optional[Union[List[str], constr(regex=r"^.+;$")]] = None
    """
    Desktop environments that should display this item.
    It'll check against ``$XDG_CURRENT_DESKTOP``.
    """
    PrefersNonDefaultGPU: Optional[bool] = None
    "Hint that the app prefers to be run on a more powerful discrete GPU if available."
    StartupNotify: Optional[bool] = None
    """
    Advanced. See `Startup Notification spec <startup-notification-spec>`__.

    .. startup-notification-spec: _https://www.freedesktop.org/wiki/Specifications/startup-notification-spec/
    """
    StartupWMClass: Optional[str] = None
    """
    Advanced. See `Startup Notification spec <startup-notification-spec>`__.

    .. startup-notification-spec: _https://www.freedesktop.org/wiki/Specifications/startup-notification-spec/
    """
    TryExec: Optional[str] = None
    """
    Filename or absolute path to an executable file on disk used to
    determine if the program is actually installed and can be run. If the test
    fails, the shortcut might be ignored / hidden.
    """
    glob_patterns: Optional[Dict[str, constr(regex=r".*\*.*")]] = None
    """
    Map of custom MIME types to their corresponding glob patterns (e.g. ``*.txt``).
    Only needed if you define custom MIME types in ``MimeType``.
    """


class MacOS(BasePlatformSpecific):
    """
    Mac-specific instructions. Check these URLs for more info:

    - ``CF*`` keys: see `Core Foundation Keys <cf-keys>`_
    - ``LS*`` keys: see `Launch Services Keys <ls-keys>`_
    - ``entitlements``: see `entitlements docs <entitlements-docs>`_

    .. _cf-keys: https://developer.apple.com/library/archive/documentation/General/Reference/InfoPlistKeyReference/Articles/CoreFoundationKeys.html
    .. _ls-keys: https://developer.apple.com/library/archive/documentation/General/Reference/InfoPlistKeyReference/Articles/LaunchServicesKeys.html
    .. _entitlements-docs: https://developer.apple.com/documentation/bundleresources/entitlements.
    """

    class CFBundleURLTypesModel(BaseModel):
        "Describes a URL scheme associated with the app."
        CFBundleTypeRole: Optional[Literal["Editor", "Viewer", "Shell", "None"]] = None
        "This key specifies the app's role with respect to the URL."
        CFBundleURLSchemes: List[str] = ...
        "URL schemes / protocols handled by this type (e.g. 'mailto').",
        CFBundleURLName: Optional[str] = None
        "Abstract name for this URL type. Uniqueness recommended.",
        CFBundleURLIconFile: Optional[str] = None
        "Name of the icon image file (minus the .icns extension).",

    class CFBundleDocumentTypesModel(BaseModel):
        "Describes a document type associated with the app."
        CFBundleTypeIconFile: Optional[str] = None
        "Name of the icon image file (minus the .icns extension).",
        CFBundleTypeName: str = ...
        "Abstract name for this document type. Uniqueness recommended.",
        CFBundleTypeRole: Optional[Literal["Editor", "Viewer", "Shell", "None"]] = None
        "This key specifies the app's role with respect to the type."
        LSItemContentTypes: List[str] = ...
        """
        List of UTI strings defining a supported file type; e.g. for
        PNG files, use 'public.png'. See `UTI Reference <uti-reference>`_
        for more info about the system-defined UTIs. Custom UTIs can be
        defined via 'UTExportedTypeDeclarations'. UTIs defined by other
        apps (not the system) need to be imported via 'UTImportedTypeDeclarations'.

        See `Fun with UTIs <fun-with-utis>`_ for more info.

        .. _uti-reference: https://developer.apple.com/library/archive/documentation/Miscellaneous/Reference/UTIRef/Articles/System-DeclaredUniformTypeIdentifiers.html
        .. _fun-with-utis: https://www.cocoanetics.com/2012/09/fun-with-uti/
        """
        LSHandlerRank: Literal["Owner", "Default", "Alternate"] = ...
        """
        Determines how Launch Services ranks this app among the apps
        that declare themselves editors or viewers of files of this type.
        """

    class UTTypeDeclarationModel(BaseModel):
        UTTypeConformsTo: List[str] = ...
        "The Uniform Type Identifier types that this type conforms to."
        UTTypeDescription: Optional[str] = None
        "A description for this type."
        UTTypeIconFile: Optional[str] = None
        "The bundle icon resource to associate with this type."
        UTTypeIdentifier: str = ...
        "The Uniform Type Identifier to assign to this type."
        UTTypeReferenceURL: Optional[str] = None
        "The webpage for a reference document that describes this type."
        UTTypeTagSpecification: Dict[str, List[str]] = ...
        "A dictionary defining one or more equivalent type identifiers."

    CFBundleDisplayName: Optional[str] = None
    """
    Display name of the bundle, visible to users and used by Siri. If
    not provided, 'menuinst' will use the 'name' field.
    """
    CFBundleIdentifier: Optional[constr(regex=r"^[A-z0-9\-\.]+$")] = None
    """
    Unique identifier for the shortcut. Typically uses a reverse-DNS format.
    If not provided, 'menuinst' will generate one from the 'name' field.
    """
    CFBundleName: Optional[constr(max_length=16)] = None
    """
    Short name of the bundle. May be used if ``CFBundleDisplayName`` is
    absent. If not provided, 'menuinst' will generate one from the 'name' field.
    """
    CFBundleSpokenName: Optional[str] = None
    """
    Suitable replacement for text-to-speech operations on the app.
    For example, 'my app one two three' instead of 'MyApp123'.
    """
    CFBundleVersion: Optional[constr(regex=r"^\S+$")] = None
    """
    Build version number for the bundle. In the context of 'menuinst'
    this can be used to signal a new version of the menu item for the same
    application version.
    """
    CFBundleURLTypes: Optional[List[CFBundleURLTypesModel]] = None
    """
    URL types supported by this app. Requires setting `event_handler` too.
    Note this feature requires macOS 10.15+.
    """
    CFBundleDocumentTypes: Optional[List[CFBundleDocumentTypesModel]] = None
    """
    Document types supported by this app. Requires setting `event_handler` too.
    Requires macOS 10.15+.
    """
    LSApplicationCategoryType: Optional[constr(regex=r"^public\.app-category\.\S+$")] = None
    "The App Store uses this string to determine the appropriate categorization."
    LSBackgroundOnly: Optional[bool] = None
    "Specifies whether this app runs only in the background."
    LSEnvironment: Optional[Dict[str, str]] = None
    "List of key-value pairs used to define environment variables.",
    LSMinimumSystemVersion: Optional[constr(regex=r"^\d+\.\d+\.\d+$")] = None
    """
    Minimum version of macOS required for this app to run, as ``x.y.z``.
    For example, for macOS v10.4 and later, use ``10.4.0``.
    """
    LSMultipleInstancesProhibited: Optional[bool] = None
    "Whether an app is prohibited from running simultaneously in multiple user sessions."
    LSRequiresNativeExecution: Optional[bool] = None
    """
    If true, prevent a universal binary from being run under
    Rosetta emulation on an Intel-based Mac.
    """
    UTExportedTypeDeclarations: Optional[List[UTTypeDeclarationModel]] = None
    "The uniform type identifiers owned and exported by the app."
    UTImportedTypeDeclarations: Optional[List[UTTypeDeclarationModel]] = None
    "The uniform type identifiers inherently supported, but not owned, by the app."
    entitlements: Optional[List[constr(regex=r"[a-z0-9\.\-]+")]] = None
    """
    List of permissions to request for the launched application.
    See `the entitlements docs <entitlements>`__ for a full list of possible values.

    .. entitlements: https://developer.apple.com/documentation/bundleresources/entitlements
    """
    link_in_bundle: Optional[
        Dict[constr(min_length=1), constr(regex=r"^(?!\/)(?!\.\./).*")]
    ] = None
    """
    Paths that should be symlinked into the shortcut app bundle.
    It takes a mapping of source to destination paths. Destination paths must be
    relative. Placeholder ``{{ MENU_ITEM_LOCATION }}`` can be useful.
    """
    event_handler: Optional[constr(min_length=1)] = None
    """
    Required shell script logic to handle opened URL payloads.
    """


class Platforms(BaseModel):
    """
    Platform specific options.

    Note each of these fields supports the same keys as the top-level :class:`MenuItem`
    (sans ``platforms`` itself), in case overrides are needed.
    """

    linux: Optional[Linux]
    "Options for Linux. See :class:`Linux` model for details."
    osx: Optional[MacOS]
    "Options for macOS. See :class:`MacOS` model for details."
    win: Optional[Windows]
    "Options for Windows. See :class:`Windows` model for details."


class MenuItem(BaseModel):
    "Instructions to create a menu item across operating systems."

    name: constr(min_length=1) = ...
    "The name of the menu item."
    description: str = ...
    "A longer description of the menu item. Shown on popup messages."
    command: conlist(str, min_items=1) = ...
    """
    Command to run with the menu item, expressed as a
    list of strings where each string is an argument.
    """
    icon: Optional[constr(min_length=1)] = None
    "Path to the file representing or containing the icon."
    precommand: Optional[constr(min_length=1)] = None
    """
    (Simple, preferrably single-line) logic to run before the command is run.
    Runs before the environment is activated, if that applies.
    """
    precreate: Optional[constr(min_length=1)] = None
    "(Simple, preferrably single-line) logic to run before the shortcut is created."
    working_dir: Optional[constr(min_length=1)] = None
    """
    Working directory for the running process.
    Defaults to user directory on each platform.
    """
    activate: Optional[bool] = True
    "Whether to activate the target environment before running ``command``."
    terminal: Optional[bool] = False
    """
    Whether run the program in a terminal/console or not.
    On Windows, it only has an effect if ``activate`` is true.
    On MacOS, the application will ignore command-line arguments.
    """
    platforms: Platforms
    "Platform-specific options. Presence of a platform field enables menu items in that platform."


class MenuInstSchema(BaseModel):
    "Metadata required to create menu items across operating systems with ``menuinst``."

    id_: Literal["https://schemas.conda.io/menuinst-1.schema.json"] = Field(
        ...,
        description="Version of the menuinst schema.",
        alias="$id",
    )
    schema_: Literal["https://json-schema.org/draft-07/schema"] = Field(
        ...,
        description="Standard of the JSON schema we adhere to.",
        alias="$schema",
    )
    menu_name: constr(min_length=1) = ...
    "Name for the category containing the items specified in ``menu_items``."
    menu_items: conlist(MenuItem, min_items=1) = ...
    "List of menu entries to create across main desktop systems."


def dump_schema_to_json(write=True):
    if write:
        here = Path(__file__).parent
        schema = MenuInstSchema.schema_json(indent=2)
        print(schema)
        with open(here / "data" / "menuinst.schema.json", "w") as f:
            f.write(schema)
            f.write("\n")
    return MenuInstSchema.schema()


def dump_default_to_json(write=True):
    here = Path(__file__).parent
    default_item = MenuItem(
        name="REQUIRED", description="REQUIRED", command=["REQUIRED"], platforms={}
    ).dict()
    default_item["platforms"] = {
        "win": Windows().dict(),
        "osx": MacOS().dict(),
        "linux": Linux().dict(),
    }
    default = MenuInstSchema(
        menu_name="REQUIRED",
        menu_items=[default_item],
        **{
            "$id": "https://schemas.conda.io/menuinst-1.schema.json",
            "$schema": "https://json-schema.org/draft-07/schema",
        },
    ).dict()
    for platform_value in default["menu_items"][0]["platforms"].values():
        for key in list(platform_value.keys()):
            if key in MenuItem.__fields__:
                platform_value.pop(key)
    if write:
        pprint(default)
        with open(here / "data" / "menuinst.default.json", "w") as f:
            json.dump(default, f, indent=2)
            f.write("\n")
    return default


def validate(metadata_or_path):
    if isinstance(metadata_or_path, (str, Path)):
        with open(metadata_or_path) as f:
            metadata = json.load(f)
    else:
        metadata = metadata_or_path
    return MenuInstSchema(**metadata)


if __name__ == "__main__":
    dump_schema_to_json()
    dump_default_to_json()



# Copyright (c) 2008-2011 by Enthought, Inc.
# Copyright (c) 2013-2015 Continuum Analytics, Inc.
# All rights reserved.

from __future__ import absolute_import

import json
import logging
import sys
from os.path import abspath, basename, exists, join

try:
    from .._version import __version__
except ImportError:
    __version__ = "dev"

from ..utils import DEFAULT_BASE_PREFIX, DEFAULT_PREFIX

if sys.platform == 'win32':
    from ..platforms.win_utils.win_elevate import isUserAdmin, runAsAdmin
    from .win32 import Menu, ShortCut


def _install(path, remove=False, prefix=None, mode=None, root_prefix=None):
    prefix = prefix or DEFAULT_PREFIX
    root_prefix = root_prefix or DEFAULT_BASE_PREFIX
    if abspath(prefix) == abspath(root_prefix):
        env_name = None
    else:
        env_name = basename(prefix)

    data = json.load(open(path))
    try:
        menu_name = data['menu_name']
    except KeyError:
        menu_name = 'Python-%d.%d' % sys.version_info[:2]

    shortcuts = data['menu_items']
    m = Menu(menu_name, prefix=prefix, env_name=env_name, mode=mode, root_prefix=root_prefix)
    if remove:
        for sc in shortcuts:
            ShortCut(m, sc).remove()
        m.remove()
    else:
        m.create()
        for sc in shortcuts:
            ShortCut(m, sc).create()


def install(path, remove=False, prefix=None, recursing=False, root_prefix=None):
    """
    Install Menu and shortcuts

    # Specifying `root_prefix` is used with conda-standalone, because we can't use
    # `sys.prefix`, therefore we need to specify it
    """
    prefix = prefix or DEFAULT_PREFIX
    root_prefix = root_prefix or DEFAULT_BASE_PREFIX
    if not sys.platform == 'win32':
        raise RuntimeError("menuinst._legacy is only supported on Windows.")

    # this root_prefix is intentional.  We want to reflect the state of the root installation.

    if not exists(join(root_prefix, '.nonadmin')):
        if isUserAdmin():
            _install(path, remove, prefix, mode='system', root_prefix=root_prefix)
        else:
            retcode = 1
            try:
                if not recursing:
                    retcode = runAsAdmin(
                        [
                            join(root_prefix, 'python'),
                            '-c',
                            "import menuinst; menuinst.install(%r, %r, %r, %r, %r)"
                            % (path, bool(remove), prefix, True, root_prefix),
                        ]
                    )
            except OSError:
                pass

            if retcode != 0:
                logging.warn(
                    "Insufficient permissions to write menu folder.  "
                    "Falling back to user location"
                )
                _install(path, remove, prefix, mode='user', root_prefix=root_prefix)
    else:
        _install(path, remove, prefix, mode='user', root_prefix=root_prefix)



from __future__ import annotations

import sys
import threading
from types import ModuleType
from typing import Any
from typing import Callable
from typing import cast
from typing import Iterable

from ._importing import _py_abspath
from ._importing import importobj
from ._syncronized import _synchronized
from menuinst._vendor.apipkg import AliasModule


class ApiModule(ModuleType):
    """the magical lazy-loading module standing"""

    def __docget(self) -> str | None:
        try:
            return self.__doc
        except AttributeError:
            if "__doc__" in self.__map__:
                return cast(str, self.__makeattr("__doc__"))
            else:
                return None

    def __docset(self, value: str) -> None:
        self.__doc = value

    __doc__ = property(__docget, __docset)  # type: ignore
    __map__: dict[str, tuple[str, str]]

    def __init__(
        self,
        name: str,
        importspec: dict[str, Any],
        implprefix: str | None = None,
        attr: dict[str, Any] | None = None,
    ) -> None:
        super().__init__(name)
        self.__name__ = name
        self.__all__ = [x for x in importspec if x != "__onfirstaccess__"]
        self.__map__ = {}
        self.__implprefix__ = implprefix or name
        if attr:
            for name, val in attr.items():
                setattr(self, name, val)
        for name, importspec in importspec.items():
            if isinstance(importspec, dict):
                subname = f"{self.__name__}.{name}"
                apimod = ApiModule(subname, importspec, implprefix)
                sys.modules[subname] = apimod
                setattr(self, name, apimod)
            else:
                parts = importspec.split(":")
                modpath = parts.pop(0)
                attrname = parts and parts[0] or ""
                if modpath[0] == ".":
                    modpath = implprefix + modpath

                if not attrname:
                    subname = f"{self.__name__}.{name}"
                    apimod = AliasModule(subname, modpath)
                    sys.modules[subname] = apimod
                    if "." not in name:
                        setattr(self, name, apimod)
                else:
                    self.__map__[name] = (modpath, attrname)

    def __repr__(self):
        repr_list = [f"<ApiModule {self.__name__!r}"]
        if hasattr(self, "__version__"):
            repr_list.append(f" version={self.__version__!r}")
        if hasattr(self, "__file__"):
            repr_list.append(f" from {self.__file__!r}")
        repr_list.append(">")
        return "".join(repr_list)

    @_synchronized
    def __makeattr(self, name, isgetattr=False):
        """lazily compute value for name or raise AttributeError if unknown."""
        target = None
        if "__onfirstaccess__" in self.__map__:
            target = self.__map__.pop("__onfirstaccess__")
            fn = cast(Callable[[], None], importobj(*target))
            fn()
        try:
            modpath, attrname = self.__map__[name]
        except KeyError:
            # __getattr__ is called when the attribute does not exist, but it may have
            # been set by the onfirstaccess call above. Infinite recursion is not
            # possible as __onfirstaccess__ is removed before the call (unless the call
            # adds __onfirstaccess__ to __map__ explicitly, which is not our problem)
            if target is not None and name != "__onfirstaccess__":
                return getattr(self, name)
            # Attribute may also have been set during a concurrent call to __getattr__
            # which executed after this call was already waiting on the lock. Check
            # for a recently set attribute while avoiding infinite recursion:
            # * Don't call __getattribute__ if __makeattr was called from a data
            #   descriptor such as the __doc__ or __dict__ properties, since data
            #   descriptors are called as part of object.__getattribute__
            # * Only call __getattribute__ if there is a possibility something has set
            #   the attribute we're looking for since __getattr__ was called
            if threading is not None and isgetattr:
                return super().__getattribute__(name)
            raise AttributeError(name)
        else:
            result = importobj(modpath, attrname)
            setattr(self, name, result)
            # in a recursive-import situation a double-del can happen
            self.__map__.pop(name, None)
            return result

    def __getattr__(self, name):
        return self.__makeattr(name, isgetattr=True)

    def __dir__(self) -> Iterable[str]:
        yield from super().__dir__()
        yield from self.__map__

    @property
    def __dict__(self) -> dict[str, Any]:  # type: ignore
        # force all the content of the module
        # to be loaded when __dict__ is read
        dictdescr = ModuleType.__dict__["__dict__"]  # type: ignore
        ns: dict[str, Any] = dictdescr.__get__(self)
        if ns is not None:
            hasattr(self, "some")
            for name in self.__all__:
                try:
                    self.__makeattr(name)
                except AttributeError:
                    pass
        return ns


_PRESERVED_MODULE_ATTRS = {
    "__file__",
    "__version__",
    "__loader__",
    "__path__",
    "__package__",
    "__doc__",
    "__spec__",
    "__dict__",
}


def _initpkg(mod: ModuleType | None, pkgname, exportdefs, attr=None) -> ApiModule:
    """Helper for initpkg.

    Python 3.3+ uses finer grained locking for imports, and checks sys.modules before
    acquiring the lock to avoid the overhead of the fine-grained locking. This
    introduces a race condition when a module is imported by multiple threads
    concurrently - some threads will see the initial module and some the replacement
    ApiModule. We avoid this by updating the existing module in-place.

    """
    if mod is None:
        d = {"__file__": None, "__spec__": None}
        d.update(attr)
        mod = ApiModule(pkgname, exportdefs, implprefix=pkgname, attr=d)
        sys.modules[pkgname] = mod
        return mod
    else:
        f = getattr(mod, "__file__", None)
        if f:
            f = _py_abspath(f)
        mod.__file__ = f
        if hasattr(mod, "__path__"):
            mod.__path__ = [_py_abspath(p) for p in mod.__path__]
        if "__doc__" in exportdefs and hasattr(mod, "__doc__"):
            del mod.__doc__
        for name in dir(mod):
            if name not in _PRESERVED_MODULE_ATTRS:
                delattr(mod, name)

        # Updating class of existing module as per importlib.util.LazyLoader
        mod.__class__ = ApiModule
        apimod = cast(ApiModule, mod)
        ApiModule.__init__(apimod, pkgname, exportdefs, implprefix=pkgname, attr=attr)
        return apimod
